/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <map>
#include <pan/general/debug.h>
#include <pan/general/file-util.h>
#include <pan/general/foreach.h>
#include <pan/general/i18n.h>
#include <pan/tasks/task-article.h>
#include <pan/tasks/task-groups.h>
#include <pan/tasks/task-xover.h>
#include <pan/tasks/nzb.h>
#include <pan/icons/pan-pixbufs.h>
#include "actions.h"
#include "body-pane.h"
#include "dl-headers-ui.h"
#include "group-pane.h"
#include "header-pane.h"
#include "license.h"
#include "log-ui.h"
#include "gui.h"
#include "pad.h"
#include "pan.ui.h"
#include "progress-view.h"
#include "post-ui.h"
#include "score-add-ui.h"
#include "score-view-ui.h"
#include "server-ui.h"
#include "task-pane.h"
#include "url.h"

using namespace pan;

namespace
{
  const int VIEW_QTY = 3;

  GtkWindow* get_window (GtkWidget* w)
  {
    return GTK_WINDOW (gtk_widget_get_toplevel (w));
  }

  void
  parent_set_cb (GtkWidget * widget, GtkObject * old_parent, gpointer ui_manager_g)
  {
    GtkWidget * toplevel = gtk_widget_get_toplevel (widget);
    if (GTK_IS_WINDOW (toplevel))
    {
      GtkUIManager * ui_manager = static_cast<GtkUIManager*>(ui_manager_g);
      gtk_window_add_accel_group (GTK_WINDOW(toplevel),
                                  gtk_ui_manager_get_accel_group (ui_manager));
    }
  }
}

void
GUI :: add_widget (GtkUIManager * merge,
                   GtkWidget    * widget,
                   gpointer       gui_g)
{
  const char * name (gtk_widget_get_name (widget));
  GUI * self (static_cast<GUI*>(gui_g));

  if (name && strstr(name,"main-window-")==name)
  {
    if (!GTK_IS_TOOLBAR (widget))
    {
      gtk_box_pack_start (GTK_BOX(self->_menu_vbox), widget, FALSE, FALSE, 0);
    }
    else
    {
      GtkWidget * handle_box = gtk_handle_box_new ();
      gtk_widget_show (handle_box);
      gtk_container_add (GTK_CONTAINER (handle_box), widget);
      gtk_box_pack_start (GTK_BOX(self->_menu_vbox), handle_box, FALSE, FALSE, 0);
      g_signal_connect_swapped (widget, "destroy",
                                G_CALLBACK (gtk_widget_destroy), handle_box);
    }
    gtk_widget_show (widget);
  }
}

void GUI :: show_event_log_cb (GtkWidget * w, gpointer gui_gpointer)
{
  static_cast<GUI*>(gui_gpointer)->activate_action ("show-log-dialog");
}

void GUI :: show_task_window_cb (GtkWidget * w, gpointer gui_gpointer)
{
  static_cast<GUI*>(gui_gpointer)->activate_action ("show-task-window");
}

GUI :: GUI (Data& data, Queue& queue, ArticleCache& cache, int width, int height):
  _data (data),
  _queue (queue),
  _cache (cache),
  _root (gtk_vbox_new (FALSE, 0)),
  _menu_vbox (gtk_vbox_new (FALSE, 0)),
  _group_pane (0),
  _header_pane (0),
  _body_pane (0),
  _ui_manager (gtk_ui_manager_new ()),
  _info_image (gtk_image_new_from_stock (GTK_STOCK_DIALOG_INFO, GTK_ICON_SIZE_BUTTON)),
  _error_image (gtk_image_new_from_stock (GTK_STOCK_DIALOG_ERROR, GTK_ICON_SIZE_BUTTON)),
  _connection_size_label (0),
  _queue_size_label (0),
  _queue_size_button (0),
  _taskbar (0),
  _ttips (gtk_tooltips_new ())
{
  char * filename = g_build_filename (file::get_pan_home().c_str(), "pan.ui", NULL);
#if 0
  GError * gerr (0);
  gtk_ui_manager_add_ui_from_file (_ui_manager, filename, &gerr);
  if (gerr)  {
    Log::add_err (gerr->message);
    g_clear_error (&gerr);
  }
#else
  if (!gtk_ui_manager_add_ui_from_file (_ui_manager, filename, NULL))
    gtk_ui_manager_add_ui_from_string (_ui_manager, fallback_ui_file, -1, NULL);
#endif
  g_free (filename);
  g_signal_connect (_ui_manager, "add_widget", G_CALLBACK (add_widget), this);
  add_actions (this, _ui_manager);
  g_signal_connect (_root, "parent-set", G_CALLBACK(parent_set_cb), _ui_manager);
  gtk_box_pack_start (GTK_BOX(_root), _menu_vbox, FALSE, FALSE, 0);

  _group_pane = new GroupPane (*this, data);
  _header_pane = new HeaderPane (*this, data, _cache);
  _body_pane = new BodyPane (*this, data, _cache);

  // workarea
  _workarea_bin = gtk_event_box_new ();
  GtkWidget * workarea = gtk_hpaned_new ();
  gtk_paned_pack1 (GTK_PANED(workarea), _group_pane->root(), true, true);
    GtkWidget * v = gtk_vpaned_new ();
    gtk_paned_pack1 (GTK_PANED(v), _header_pane->root(), true, true);
    gtk_paned_pack2 (GTK_PANED(v), _body_pane->root(), true, true);
  gtk_paned_pack2 (GTK_PANED(workarea), v, true, true);
  gtk_paned_pack2 (GTK_PANED(workarea), _header_pane->root(), true, true);
  gtk_paned_set_position (GTK_PANED(workarea), 250);
  gtk_container_add (GTK_CONTAINER(_workarea_bin), workarea);
  gtk_box_pack_start (GTK_BOX(_root), _workarea_bin, TRUE, TRUE, 0);

  /**
  ***  Status Bar
  **/

  GtkWidget * w = gtk_event_box_new ();
  gtk_widget_set_size_request (w, -1, PAD_SMALL);
  gtk_box_pack_start (GTK_BOX(_root), w, false, false, 0);

  GtkWidget * status_bar (gtk_hbox_new (FALSE, PAD_SMALL));

  // connection status
  w = _connection_size_label = gtk_label_new (NULL);
  gtk_misc_set_padding (GTK_MISC(w), PAD, 0);
  GtkWidget * frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_IN);
  gtk_container_add (GTK_CONTAINER(frame), w);
  gtk_box_pack_start (GTK_BOX(status_bar), frame, FALSE, FALSE, 0);

  // queue
  _queue_size_label = gtk_label_new (NULL);
  w = _queue_size_button = gtk_button_new();
  gtk_tooltips_set_tip (GTK_TOOLTIPS(_ttips), w, _("Open the Task Manager"), NULL);
  gtk_button_set_relief (GTK_BUTTON(w), GTK_RELIEF_NONE);
  g_signal_connect (GTK_OBJECT(w), "clicked", G_CALLBACK(show_task_window_cb), this);
  gtk_container_add (GTK_CONTAINER(w), _queue_size_label);
  frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_IN);
  gtk_container_add (GTK_CONTAINER(frame), w);
  gtk_box_pack_start (GTK_BOX(status_bar), frame, FALSE, FALSE, 0);

  // status item views
  _taskbar = gtk_table_new (1, VIEW_QTY, TRUE);
  for (int i=0; i<VIEW_QTY; ++i) {
    ProgressView * v = new ProgressView ();
    gtk_table_attach (GTK_TABLE(_taskbar), v->root(), i, i+1, 0, 1, (GtkAttachOptions)~0, (GtkAttachOptions)~0, 0, 0);
    _views.push_back (v);
  }
  gtk_box_pack_start (GTK_BOX(status_bar), _taskbar, true, true, 0);
  gtk_widget_show_all (status_bar);

  // status 
  w = _event_log_button = gtk_button_new ();
  gtk_tooltips_set_tip (GTK_TOOLTIPS(_ttips), w, _("Open the Event Log"), NULL);
  gtk_button_set_relief (GTK_BUTTON(w), GTK_RELIEF_NONE);
//  frame = gtk_frame_new (NULL);
//  gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_IN);
//  gtk_container_add (GTK_CONTAINER(frame), w);
  //gtk_box_pack_start (GTK_BOX(status_bar), frame, false, false, 0);
  gtk_box_pack_start (GTK_BOX(status_bar), w, false, false, 0);
  gtk_container_add (GTK_CONTAINER(w), _info_image);
  g_signal_connect (w, "clicked", G_CALLBACK(show_event_log_cb), this);

  gtk_box_pack_start (GTK_BOX(_root), status_bar, false, false, 0);

  upkeep_tag = g_timeout_add (3000, upkeep_timer_cb, this);


#if 0
        /* status icons */
        for (i=0; i<STATUS_QTY; ++i) {
                _status_icons[i].pixbuf = gdk_pixbuf_new_from_inline (-1, _status_icons[i].inline_txt,
                                                                      FALSE, NULL);
                gdk_pixbuf_render_pixmap_and_mask_for_colormap (_status_icons[i].pixbuf, cmap,
                                                                &_status_icons[i].pixmap,
                                                                &_status_icons[i].bitmap, 128);
        }

        /* status button */
        w = gtk_button_new ();
        _status_image = gtk_image_new ();
        gtk_container_add (GTK_CONTAINER(w), _status_image);
#endif

  _queue.add_listener (this);
  Log::get().add_listener (this);

  gtk_widget_ref (_info_image);
  gtk_widget_ref (_error_image);
  gtk_widget_ref (_group_pane->root());
  gtk_widget_ref (_header_pane->root());
  gtk_widget_ref (_body_pane->root());

  // update the connections label
  upkeep_timer_cb (this);
  // update the queue label
  set_queue_size_label (0, 0);
}

GUI :: ~GUI ()
{
  _queue.remove_listener (this);
  Log::get().remove_listener (this);
  g_source_remove (upkeep_tag);

  gtk_widget_unref (_body_pane->root());
  gtk_widget_unref (_header_pane->root());
  gtk_widget_unref (_group_pane->root());
  gtk_widget_unref (_error_image);
  gtk_widget_unref (_info_image);

  delete _header_pane;
  delete _group_pane;
  delete _body_pane;
  for (size_t i(0), size(_views.size()); i!=size; ++i)
    delete _views[i];
}

/***
****
***/

void
GUI :: watch_cursor_on ()
{
  GdkCursor * cursor = gdk_cursor_new (GDK_WATCH);
  gdk_window_set_cursor (_root->window, cursor);
  gdk_cursor_unref (cursor);
  while (gtk_events_pending ())
    gtk_main_iteration ();
}

void
GUI :: watch_cursor_off ()
{
  gdk_window_set_cursor (_root->window, NULL);
}

/***
****
***/

namespace
{
  typedef std::map < std::string, GtkAction* > key_to_action_t;

  key_to_action_t key_to_action;

  void ensure_action_map_loaded (GtkUIManager * uim)
  {
    if (!key_to_action.empty())
      return;
    for (GList * l=gtk_ui_manager_get_action_groups(uim); l!=0; l=l->next) {
      GtkActionGroup * action_group = GTK_ACTION_GROUP(l->data);
      for (GList * al=gtk_action_group_list_actions(action_group); al!=0; al=al->next) {
        GtkAction * action = GTK_ACTION (al->data);
        const std::string name (gtk_action_get_name (action));
        key_to_action[name] = action;
      }
    }
  }

  GtkAction * get_action (const char * name) {
    key_to_action_t::iterator it = key_to_action.find (name);
    assert (it != key_to_action.end());
    return it->second;
  }
}

bool
GUI :: is_action_active (const char *key) const
{
  ensure_action_map_loaded (_ui_manager);
  return gtk_toggle_action_get_active (GTK_TOGGLE_ACTION(get_action(key)));
}

void
GUI :: activate_action (const char * key) const
{
  ensure_action_map_loaded (_ui_manager);
  gtk_action_activate (get_action(key));
}

void
GUI :: toggle_action (const char * key, bool b) const
{
  ensure_action_map_loaded (_ui_manager);
  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(get_action(key)), b);
}

GtkWidget*
GUI :: get_action_widget (const char * key) const
{
  return gtk_ui_manager_get_widget (_ui_manager, key);
}

namespace
{
  gboolean focus_in_cb (GtkWidget     * w,
                        GdkEventFocus * unused,
                        gpointer        accel_group_g)
  {
    GtkAccelGroup * accel_group = static_cast<GtkAccelGroup*>(accel_group_g);
    gtk_window_remove_accel_group (get_window(w), accel_group);
    return false;
  }

#if 0
  g_object_ref (G_OBJECT(accel_group));
  g_object_set_data_full (G_OBJECT(_entry), "accel-group", accel_group, g_object_unref);
#endif


  gboolean focus_out_cb (GtkWidget     * w,
                         GdkEventFocus * unused,
                         gpointer        accel_group_g)
  {
    GtkAccelGroup * accel_group = static_cast<GtkAccelGroup*>(accel_group_g);
    gtk_window_add_accel_group (get_window(w), accel_group);
    return false;
  }
}

void
GUI :: disable_accelerators_when_focused (GtkWidget * w) const
{
  GtkAccelGroup * accel_group = gtk_ui_manager_get_accel_group (_ui_manager);
  g_signal_connect (w, "focus-in-event", G_CALLBACK(focus_in_cb), accel_group);
  g_signal_connect (w, "focus-out-event", G_CALLBACK(focus_out_cb), accel_group);
}

/***
****  PanUI
***/

namespace
{
  std::string get_save_attachments_path_from_user (GtkWindow * parent)
  {
    static std::string prev_path;
    if (prev_path.empty())
      prev_path = g_get_home_dir ();
    GtkWidget * w = gtk_file_chooser_dialog_new (_("Save Attachments"), parent,
                                                 GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                                 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                                 GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
                                                 NULL);
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(w), prev_path.c_str());
    const int response (gtk_dialog_run (GTK_DIALOG(w)));
    std::string path;
    if (response == GTK_RESPONSE_ACCEPT) {
      char * tmp = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (w));
      prev_path = path = tmp;
      g_free (tmp);
    }

    gtk_widget_destroy( w);
    return path;
  }
}

void GUI :: do_save_attachments ()
{
  std::string path;
  std::set<const Article*> articles (_header_pane->get_full_selection ());
  if (!articles.empty())
    path = get_save_attachments_path_from_user (get_window(_root));
  if (!path.empty()) {
    std::vector<Task*> tasks;
    foreach (std::set<const Article*>, articles, it)
      tasks.push_back (new TaskArticle (_data, **it, _cache, _data, NULL, path));
    _queue.add_tasks (tasks);
  }
}

void GUI :: do_print ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_cancel_latest_task ()
{
  _queue.remove_latest_task ();
}
void GUI :: do_import_tasks ()
{
  // get a list of files to import
  GtkWidget * dialog = gtk_file_chooser_dialog_new (
    _("Import NZB File(s)"), get_window(_root),
    GTK_FILE_CHOOSER_ACTION_OPEN,
    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
    GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
    NULL);
  gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER(dialog), true);
  typedef std::vector<std::string> strings_t;
  strings_t filenames;
  if (gtk_dialog_run (GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
    GSList * tmp = gtk_file_chooser_get_filenames (GTK_FILE_CHOOSER(dialog));
    for (GSList * l(tmp); l!=0; l=l->next) {
      filenames.push_back ((char*) l->data);
      g_free (l->data);
    }
    g_slist_free (tmp);
  }
  gtk_widget_destroy (dialog);

  // if we're importing files, build the tasks...
  std::vector<Task*> tasks;
  if (!filenames.empty()) {
    const std::string path (get_save_attachments_path_from_user (get_window(_root)));
    if (!path.empty())
      foreach_const (strings_t, filenames, it)
        NZB :: tasks_from_nzb_file (*it, path, _cache, _data, _data, tasks);
  }

  if (!tasks.empty())
    _queue.add_tasks (tasks);
}

void GUI :: do_show_task_window ()
{
  TaskPane * task_pane = new TaskPane (_queue, _root);
  gtk_widget_show (task_pane->root());
}

namespace
{
  void set_bin_child (GtkWidget * w, GtkWidget * new_child)
  {
    GtkWidget * child (gtk_bin_get_child (GTK_BIN(w)));
    if (child != new_child)
    {
      gtk_container_remove (GTK_CONTAINER(w), child);
      gtk_container_add (GTK_CONTAINER(w), new_child);
      gtk_widget_show (new_child);
    }
  } 
}

void GUI :: on_log_entry_added (const Log::Entry& e)
{
  if (e.severity & Log::PAN_SEVERITY_ERROR)
    set_bin_child (_event_log_button, _error_image);
}

void GUI :: do_show_log_window ()
{
  gtk_widget_show (log_dialog_new (get_window (_root)));
  set_bin_child (_event_log_button, _info_image);
}
void GUI :: do_select_all_articles ()
{
  _header_pane->select_all ();
}
void GUI :: do_unselect_all_articles ()
{
  _header_pane->unselect_all ();
}
void GUI :: do_add_threads_to_selection ()
{
  _header_pane->select_threads ();
}
void GUI :: do_add_subthreads_to_selection ()
{
  _header_pane->select_subthreads ();
}
void GUI :: do_select_article_body ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_show_preferences_dialog ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}

void GUI :: do_jump_to_group_tab ()
{
  toggle_action ("tabbed-layout", true);
  GtkNotebook * n (GTK_NOTEBOOK (GTK_BIN (_workarea_bin)->child));
  gtk_notebook_set_current_page (n, 0);
}

void GUI :: do_jump_to_header_tab ()
{
  toggle_action ("tabbed-layout", true);
  GtkNotebook * n (GTK_NOTEBOOK (GTK_BIN (_workarea_bin)->child));
  gtk_notebook_set_current_page (n, 1);
}

void GUI :: do_jump_to_body_tab ()
{
  toggle_action ("tabbed-layout", true);
  GtkNotebook * n (GTK_NOTEBOOK (GTK_BIN (_workarea_bin)->child));
  gtk_notebook_set_current_page (n, 2);
}

void GUI :: do_rot13_selected_text ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}


void GUI :: on_progress_finished (Progress & p, int status)
{
  TaskArticle * ta = dynamic_cast<TaskArticle*>(&p);
  if (status==OK && ta && ta->get_article().message_id == latest_read_article.message_id)
    _body_pane->set_article (latest_read_article);
}

void GUI :: do_read_selected_article ()
{
  const Article* article (_header_pane->get_first_selected_article ());
  if (article) {
    latest_read_article = *article;
    _queue.add_task (new TaskArticle (_data, *article, _cache, _data, this));
  }
}
void GUI :: do_read_more ()
{
  if (!_body_pane->read_more ())
    activate_action ("read-next-unread-article");
}
void GUI :: do_read_less ()
{
  if (!_body_pane->read_less ())
    activate_action ("read-previous-article");
}
void GUI :: do_read_next_unread_article ()
{
  _header_pane->read_next_unread_article ();
}
void GUI :: do_read_next_article ()
{
  _header_pane->read_next_article ();
}
void GUI :: do_read_next_watched_article ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_read_next_unread_thread ()
{
  _header_pane->read_next_unread_thread ();
}
void GUI :: do_read_next_thread ()
{
  _header_pane->read_next_thread ();
}
void GUI :: do_read_previous_article ()
{
  _header_pane->read_previous_article ();
}
void GUI :: do_read_previous_thread ()
{
  _header_pane->read_previous_thread ();
}
void GUI :: do_read_parent_article ()
{
  _header_pane->read_parent_article ();
}

void GUI ::  server_list_dialog_destroyed_cb (GtkWidget * w, gpointer self)
{
  static_cast<GUI*>(self)->server_list_dialog_destroyed (w);
}

// this queues up a grouplist task for any servers that
// were added while the server list dialog was up.
void GUI :: server_list_dialog_destroyed (GtkWidget * w)
{
  quarks_t empty_servers, all_servers (_data.get_servers());
  foreach_const (quarks_t, all_servers, it) {
    quarks_t tmp;
    _data.server_get_groups (*it, tmp);
    if (tmp.empty())
      _queue.add_task (new TaskGroups (_data, *it));
  }
}

void GUI :: do_show_servers_dialog ()
{
  GtkWidget * w = server_list_dialog_new (_data, get_window(_root));
  gtk_widget_show_all (w);
  g_signal_connect (w, "destroy", G_CALLBACK(server_list_dialog_destroyed_cb), this);
}

void GUI :: do_show_score_dialog ()
{
  const Quark& group (_header_pane->get_group());
  const Article * article (0);
  if (!group.empty())
    article = _header_pane->get_first_selected_article ();
  if (article) {
    GtkWindow * window (get_window (_root));
    ScoreView * view = new ScoreView (_data, window, group, *article);
    gtk_widget_show (view->root());
  }
}

void GUI :: do_watch ()
{
  score_add (ScoreAddDialog::WATCH_SUBTHREAD);
}
void GUI :: do_ignore ()
{
  score_add (ScoreAddDialog::IGNORE_SUBTHREAD);
}
void GUI :: do_plonk ()
{
  score_add (ScoreAddDialog::PLONK);
}
void GUI :: do_show_new_score_dialog ()
{
  score_add (ScoreAddDialog::ADD);
}
void
GUI :: score_add (int mode)
{
  Quark group (_header_pane->get_group());
  if (group.empty())
    group = _group_pane->get_selection();

  Article a;
  const Article* article (_header_pane->get_first_selected_article ());
  if (article != 0)
    a = *article;

  ScoreAddDialog * d = new ScoreAddDialog (_data, _root, group, a, (ScoreAddDialog::Mode)mode);
  gtk_widget_show (d->root());
}

void GUI :: do_cancel_article ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_supersede_article ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_delete_article ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void
GUI :: do_post ()
{
  GMimeMessage * message = g_mime_message_new (false);

  // add newsgroup...
  Quark group (_header_pane->get_group ());
  if (group.empty())
    group = _group_pane->get_selection ();
  if (!group.empty())
    g_mime_message_add_header (message, "Newsgroups", group.c_str());

  // content type
  GMimePart * part = g_mime_part_new ();
  g_mime_part_set_content_type (part, g_mime_content_type_new_from_string ("text/plain; charset=UTF-8"));
  g_mime_part_set_encoding (part, GMIME_PART_ENCODING_8BIT);
  g_mime_message_set_mime_part (message, GMIME_OBJECT(part));
  g_object_unref (part);

  PostUI * post = new PostUI (_queue, _data, _data, message, get_window(_root));
  gtk_widget_show (post->root());
  g_object_unref (message);
}

void GUI :: do_followup ()
{
  GMimeMessage * message = _body_pane->create_followup ();
  if (message) {
    PostUI * post = new PostUI (_queue, _data, _data, message, get_window(_root));
    gtk_widget_show (post->root());
    g_object_unref (message);
  }
}
void GUI :: do_pan_web ()
{
  pan_url_show ("http://pan.rebelbase.com/");
}
void GUI :: do_bug_report ()
{
  pan_url_show ("http://bugzilla.gnome.org/enter_bug.cgi?product=Pan");
}
void GUI :: do_tip_jar ()
{
  pan_url_show ("http://pan.rebelbase.com/tipjar/");
}
void GUI :: do_about_pan ()
{
#if GTK_CHECK_VERSION(2,6,0)
  const gchar * authors [] = { "Charles Kerr", 0 };
  GdkPixbuf * logo = gdk_pixbuf_new_from_inline(-1, icon_pan_about_logo, 0, 0);
  GtkAboutDialog * w (GTK_ABOUT_DIALOG (gtk_about_dialog_new ()));
  gtk_about_dialog_set_name (w, _("Pan"));
  gtk_about_dialog_set_version (w, VERSION);
  gtk_about_dialog_set_comments (w, VERSION_TITLE);
  gtk_about_dialog_set_copyright (w, _("Copyright © 2002-2006 Charles Kerr"));
  gtk_about_dialog_set_website (w, "http://pan.rebelbase.com/");
  gtk_about_dialog_set_logo (w, logo);
  gtk_about_dialog_set_license (w, LICENSE);
  gtk_about_dialog_set_authors (w, authors);
  gtk_about_dialog_set_translator_credits (w, _("translator-credits"));
  gtk_widget_show_all (GTK_WIDGET(w));
  g_object_unref (logo);
#endif
}

void GUI :: do_work_online (bool b)
{
  _queue.set_online (b);
}

namespace
{
  void set_visible (GtkWidget * w, bool visible)
  {
    if (visible)
      gtk_widget_show (w);
    else
      gtk_widget_hide (w);
  }
}

namespace
{
  void remove_from_parent (GtkWidget * w)
  {
    if (w->parent != 0)
      gtk_container_remove (GTK_CONTAINER(w->parent), w);
  }
}

void GUI :: do_tabbed_layout (bool tabbed)
{
  gtk_widget_hide_all (_workarea_bin);

  GtkWidget * group_w (_group_pane->root());
  GtkWidget * header_w (_header_pane->root());
  GtkWidget * body_w (_body_pane->root());

  remove_from_parent (group_w);
  remove_from_parent (header_w);
  remove_from_parent (body_w);

  // remove workarea's current child
  GList * children = gtk_container_children (GTK_CONTAINER(_workarea_bin));
  gtk_container_remove (GTK_CONTAINER(_workarea_bin),
                        GTK_WIDGET(children->data));
  g_list_free (children);

  GtkWidget * w;
  if (tabbed) {
    w = gtk_notebook_new ();
    GtkNotebook * n (GTK_NOTEBOOK (w));
    gtk_notebook_append_page (n, group_w, gtk_label_new_with_mnemonic (_("_1. Group Pane")));
    gtk_notebook_append_page (n, header_w, gtk_label_new_with_mnemonic (_("_2. Header Pane")));
    gtk_notebook_append_page (n, body_w, gtk_label_new_with_mnemonic (_("_3. Body Pane")));
    gtk_notebook_set_tab_border (n, 2);
  } else {
    w = gtk_hpaned_new ();
      GtkWidget * v = gtk_vpaned_new ();
      gtk_paned_pack1 (GTK_PANED(v), header_w, true, true);
      gtk_paned_pack2 (GTK_PANED(v), body_w, true, true);
    gtk_paned_pack1 (GTK_PANED(w), group_w, true, true);
    gtk_paned_pack2 (GTK_PANED(w), v, true, true);
    gtk_paned_set_position (GTK_PANED(w), 250);
  }

  gtk_container_add (GTK_CONTAINER(_workarea_bin), w);
  gtk_widget_show_all (_workarea_bin);
  set_visible (group_w, is_action_active("show-group-pane"));
  set_visible (header_w, is_action_active("show-header-pane"));
  set_visible (body_w, is_action_active("show-body-pane"));
}

void GUI :: do_show_group_pane (bool b)
{
  set_visible (_group_pane->root(), b);
}
void GUI :: do_show_header_pane (bool b)
{
  set_visible (_header_pane->root(), b);
}
void GUI :: do_show_body_pane (bool b)
{
  set_visible (_body_pane->root(), b);
}
void GUI :: do_shorten_group_names (bool b)
{
  _group_pane->set_name_collapse (b);
}

void GUI :: do_wrap_article_body (bool b) { _body_pane->refresh (); }
void GUI :: do_mute_quoted_text (bool b) { _body_pane->refresh (); }
void GUI :: do_show_all_headers (bool b) { _body_pane->refresh (); }
void GUI :: do_show_smilies (bool b) { _body_pane->refresh (); }

void GUI :: do_use_monospace_font (bool b)
{
  _body_pane->set_monospace (b);
}

void GUI :: do_match_only_unread_articles (bool b) { _header_pane->refilter (); }
void GUI :: do_match_only_cached_articles (bool b) { _header_pane->refilter (); }
void GUI :: do_match_only_binary_articles (bool b) { _header_pane->refilter (); }
void GUI :: do_match_only_my_articles (bool b) { _header_pane->refilter (); }
void GUI :: do_match_on_score_state (int state) { _header_pane->refilter (); }

void GUI :: do_show_matching_articles ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_show_matching_threads ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_show_matching_subthreads ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_exit ()
{
  gtk_main_quit ();
}
void GUI :: do_read_selected_group ()
{
  const Quark group (_group_pane->get_selection ());

  // update the titlebar
  if (group.empty())
    gtk_window_set_title (get_window(_root), _("Pan"));
  else {
    char * buf = g_strdup_printf (_("Pan: %s"), group.c_str());
    gtk_window_set_title (get_window(_root), buf);
    g_free (buf);
  }

  // update the header pane
  watch_cursor_on ();
  const bool changed (_header_pane->set_group (group));
  watch_cursor_off ();

  // update the body pane
  _body_pane->clear ();

  // start a `get new'
  if (changed && _queue.is_online())
  {
    // has this group ever called `xover' for any server?
    bool group_has_had_an_xover (false);
    quarks_t servers;
    _data.group_get_servers (group, servers);
    foreach_const (quarks_t, servers, it)
      if ((group_has_had_an_xover = (_data.get_xover_high (group, *it) != 0)))
        break;

    if (group_has_had_an_xover)
      _queue.add_task (new TaskXOver (_data, group, TaskXOver::NEW));
    else // first time; get a sample
      _queue.add_task (new TaskXOver (_data, group, TaskXOver::SAMPLE));
  }
}
void GUI :: do_mark_selected_groups_read ()
{
  const quarks_t group_names (_group_pane->get_full_selection ());
  foreach (quarks_t, group_names, it)
    _data.mark_group_read (*it);
}
void GUI :: do_clear_selected_groups ()
{
  std::cerr << "FIXME " << LINE_ID << std::endl;
}
void GUI :: do_xover_selected_groups ()
{
  const quarks_t group_names (_group_pane->get_full_selection ());
  foreach (quarks_t, group_names, it)
    _queue.add_task (new TaskXOver (_data, *it, TaskXOver::NEW));
}

void GUI :: do_download_headers ()
{
  const quarks_t groups (_group_pane->get_full_selection ());
  headers_dialog (_data, _queue, groups, get_window(_root));
}

void GUI :: do_refresh_groups ()
{
  std::vector<Task*> tasks;

  const quarks_t servers (_data.get_servers ());
  foreach_const (quarks_t, servers, it)
    tasks.push_back (new TaskGroups (_data, *it));

  if (!tasks.empty())
    _queue.add_tasks (tasks);
}

void GUI :: do_subscribe_selected_groups ()
{
  const quarks_t group_names (_group_pane->get_full_selection ());
  foreach (quarks_t, group_names, it)
    _data.set_group_subscribed (*it, true);
}
void GUI :: do_unsubscribe_selected_groups ()
{
  const quarks_t group_names (_group_pane->get_full_selection ());
  foreach (quarks_t, group_names, it)
    _data.set_group_subscribed (*it, false);
}


/***
****
***/

namespace
{
  int connection_size (0);
  double KiBps (0.0);
};

void
GUI :: refresh_connection_label ()
{
  char str[128];
  char tip[128];

  // decide what to say
  if (!_queue.is_online()) {
    g_snprintf (str, sizeof(str), _("Offline"));
    if (connection_size != 0)
      g_snprintf (tip, sizeof(tip), _("Closing %d connections"), connection_size);
    else
      g_snprintf (tip, sizeof(tip), _("No Connections"));
  }
  else if (connection_size != 0) {
    g_snprintf (str, sizeof(str), "%d @ %.1f KiB/s", connection_size, KiBps);
    g_snprintf (tip, sizeof(tip), _("%d connections totaling %.1f KiB per second"),
                connection_size, KiBps);
  }
  else {
    g_snprintf (str, sizeof(str), _("No Connections"));
    g_snprintf (tip, sizeof(tip), "%s", str);
  }

  gtk_label_set_text (GTK_LABEL(_connection_size_label), str);
  gtk_tooltips_set_tip (GTK_TOOLTIPS(_ttips), _connection_size_label, tip, NULL);
}

namespace
{
  void
  timeval_diff (GTimeVal * start, GTimeVal * end, GTimeVal * diff)
  {
    diff->tv_sec = end->tv_sec - start->tv_sec;
    if (end->tv_usec < start->tv_usec)
      diff->tv_usec = 1000000ul + end->tv_usec - start->tv_usec, --diff->tv_sec;
    else
      diff->tv_usec = end->tv_usec - start->tv_usec;
  }
}

int
GUI :: upkeep_timer_cb (gpointer gui_g)
{
  static_cast<GUI*>(gui_g)->upkeep ();
  return true;
}

void
GUI :: upkeep ()
{
  KiBps = _queue.get_speed_KiBps ();

  refresh_connection_label ();
}

void
GUI :: set_queue_size_label (uint running, uint size)
{
  char str[128];
  char tip[128];

  // build the format strings
  if (size == 0u) {
    g_snprintf (str, sizeof(str), _("No Tasks"));
    *tip = '\0';
  } else {
    g_snprintf (str, sizeof(str), _("Tasks: %u/%u"), running, size);
    g_snprintf (tip, sizeof(tip), _("%u Tasks Running, %u Tasks Total"), running, size);
  }

  // update the gui
  gtk_label_set_text (GTK_LABEL(_queue_size_label), str);
  gtk_tooltips_set_tip (GTK_TOOLTIPS(_ttips), _queue_size_button, tip, NULL);
}

void
GUI :: on_queue_task_active_changed (Queue& q, Task& t, bool is_active)
{
  // update our set of active tasks
  std::list<Task*>::iterator it (std::find (_active_tasks.begin(), _active_tasks.end(), &t));
  if (is_active && it==_active_tasks.end())
    _active_tasks.push_back (&t);
  else if (!is_active && it!=_active_tasks.end())
    _active_tasks.erase (it);

  // update all the views
  int i (0);
  it = _active_tasks.begin();
  for (; it!=_active_tasks.end() && i<VIEW_QTY; ++it, ++i)
    _views[i]->set_progress (*it);
  for (; i<VIEW_QTY; ++i)
    _views[i]->set_progress (0);
}
void
GUI :: on_queue_connection_count_changed (Queue& q, int count)
{
  connection_size = count;
  refresh_connection_label ();
}
void
GUI :: on_queue_size_changed (Queue& q, int active, int total)
{
  set_queue_size_label (active, total);
}
void
GUI :: on_queue_online_changed (Queue& q, bool online)
{
  refresh_connection_label ();
}
